Imports System.IO
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports System.Globalization

Imports UvARescue.Tools


''' <summary>
''' Used for agents that obtain live sensor data from the Usarsim server.
''' </summary>
''' <remarks></remarks>
Public Class LiveDriver
    Inherits Driver

#Region " Constructor "

    Public Sub New(ByVal agent As Agent, ByVal config As TeamConfig, ByVal useImageServer As Boolean, ByVal createLogfile As Boolean) 'should also create useConnectionToUsarSim
        MyBase.New(agent)

        Me._Config = config
        Me._UseImageServer = useImageServer

        If createLogfile Then
            Me.OpenLog(String.Format(My.Settings.LogFileNameFormat, agent.Name))
        End If

    End Sub

    Private _Config As TeamConfig
    Protected ReadOnly Property Config() As TeamConfig
        Get
            Return Me._Config
        End Get
    End Property

    Private _UseImageServer As Boolean
    Protected ReadOnly Property UseImageServer() As Boolean
        Get
            Return Me._UseImageServer
        End Get
    End Property

#End Region

#Region " TcpConnection to UsarSim "

    Private _UsarSimConnection As New TcpMessagingConnection

    Public ReadOnly Property ConnectedToUsarSim() As Boolean
        Get
            Return Me._UsarSimConnection.IsConnected
        End Get
    End Property

    Protected Sub ConnectToUsarSim(ByVal host As String, ByVal port As Integer)
        If Not IsNothing(Me.Agent) AndAlso Not IsNothing(Me.Agent.Name) Then
            Console.WriteLine(String.Format("[LiveDriver]: {0} does ConnectToUsarSim", Me.Agent.Name))
        Else
            Console.WriteLine("[LiveDriver]: ConnectToUsarSim")
        End If
        Me._UsarSimConnection.Connect(host, port)
    End Sub

    Protected Sub DisconnectFromUsarSim()
        If Not IsNothing(Me.Agent) AndAlso Not IsNothing(Me.Agent.Name) Then
            Console.WriteLine(String.Format("[LiveDriver]: {0} does DisconnectFromUsarSim", Me.Agent.Name))
        Else
            Console.WriteLine("[LiveDriver]: DisconnectFromUsarSim")
        End If
        Me._UsarSimConnection.Disconnect()
    End Sub

#End Region

#Region " TcpConnection to ImageServer "

    Private _ImageServerConnection As New TcpCameraConnection

    Public ReadOnly Property ConnectedToImageServer() As Boolean
        Get
            Return Me._ImageServerConnection.IsConnected
        End Get
    End Property

    Protected Sub ConnectToImageServer(ByVal host As String, ByVal port As Integer)
        If Not IsNothing(Me.Agent) AndAlso Not IsNothing(Me.Agent.Name) Then

            If Me.Agent.Name = Me.Agent.TeamConfig.OperatorName AndAlso Me.Agent.AgentConfig.AgentType = "operator" Then
                Console.WriteLine(String.Format("[LiveDriver]: {0} with role '{1}' does not need to ConnectToImageServer", Me.Agent.Name, Me.Agent.AgentConfig.AgentType))
                Return
            End If

            Console.WriteLine(String.Format("[LiveDriver]: {0} does ConnectToImageServer", Me.Agent.Name))

        Else

            Console.WriteLine("[LiveDriver]: ConnectToImageServer")
        End If

        
        Me._ImageServerConnection.Connect(host, port)
    End Sub

    Protected Sub DisconnectFromImageServer()
        If Not IsNothing(Me.Agent) AndAlso Not IsNothing(Me.Agent.Name) Then
            Console.WriteLine(String.Format("[LiveDriver]: {0} does DisconnectToImageServer", Me.Agent.Name))
        Else
            Console.WriteLine("[LiveDriver]: DisconnectToImageServer")
        End If
        Me._ImageServerConnection.Disconnect()
    End Sub

#End Region

#Region " Driver Lifetime "

    Public Overrides Sub Start()

        With Me._Config

            Me.ConnectToUsarSim(.UsarSimHost, .UsarSimPort)

            If Me._UseImageServer Then
                Me.ConnectToImageServer(.ImageServerHost, .ImageServerPort)
            End If

            'Julian RoboCup 2008: Sleep for 3 seconds to allow robot to be spawned, before connecting
            'to new WSS.  Must be done otherwise WSS refuses connection.
            If Me._Config.WirelessServerPort = 50000 Then
                Threading.Thread.Sleep(TimeSpan.FromSeconds(3))
            End If


        End With

        If Threading.Thread.CurrentThread.Name = Nothing Then
            Threading.Thread.CurrentThread.Name = "LiveDriver"
        End If


        If Me.ConnectedToUsarSim = True Then
            Threading.Thread.CurrentThread.Priority = ThreadPriority.AboveNormal
            MyBase.Start()
        Else
            Console.WriteLine("[LiveDriver:Start] No connection to UsarSim")
            Me.Agent.Finish()
        End If

    End Sub


    Public Overrides Sub [Stop]()
        'first kill the main routine
        MyBase.Stop()

        'then disconnect 
        If Me.ConnectedToImageServer Then
            Me.DisconnectFromImageServer()
        End If
        If Me.ConnectedToUsarSim Then
            Me.DisconnectFromUsarSim()
        End If

        Me.Agent.Finish()

        Me.CloseLog()

    End Sub


    Protected Overrides Sub Run()

        System.Threading.Thread.CurrentThread.Name = Me.Agent.Name

        'See http://msdn.microsoft.com/en-us/library/system.globalization.cultureinfo.currentculture(VS.71).aspx

        If Not CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator = NumberFormatInfo.InvariantInfo.NumberDecimalSeparator Then
            'This prevents BufferedLayer overflow, due to huge laser ranges if ',' and '.' are interchanged
            Thread.CurrentThread.CurrentCulture = New CultureInfo("en-US", False)
            Console.WriteLine("[LifeDriver:Run]: CurrentCulture is now {0}.", CultureInfo.CurrentCulture.Name)
        End If


        'first, spawn the robot on the server
        Me.Agent.Spawn()

        Dim bSentCmds As Boolean
        Dim bRecvMsgs As Boolean
        Dim bRecvImgs As Boolean

        While Me.IsRunning

            'reset flags 
            bSentCmds = False
            bRecvMsgs = False
            bRecvImgs = False

            'check for pause
            Me.CheckForPause()

            If Me.Agent.UsarSimCommands.Count > 0 Then
                'send out any pending usarsim commands

                bSentCmds = True 'set the flag

                Try

                    'Failed once (after Iran Open Final, with Spawn directly). Probably Dequeue still in process when Count is tested 
                
                While Me.Agent.UsarSimCommands.Count > 0
                    Me.Log(Me.Agent.UsarSimCommands.Peek.ToString)
                    Me._UsarSimConnection.Send(Me.Agent.UsarSimCommands.Dequeue)
                    End While
                Catch ex As Exception
                    Console.WriteLine("[LiveDriver] Error occurred while dequeuing UsarSimCommands.")
                    'Console.WriteLine(ex.Message)
                End Try

            End If


            'receive any inbound messages from usarsim
            If Me._UsarSimConnection.DataAvailable Then

                bRecvMsgs = True 'set the flag

                Dim msgs As Specialized.StringCollection = Me._UsarSimConnection.ReceiveMessages(4096)
                For Each msg As String In msgs
                    'Console.WriteLine(String.Format("[msg time] = {0}", msg))
                    If Not LineParsers.ParseUsarSimLine(Me.Agent, msg) Then
                        Me.Error("Could not parse message.", msg) 'GetGeo and GetConf not supported in UT3 (yet)
                        Me.Log(msg)
                    Else
                        Me.Log(msg)
                    End If
                Next
                msgs.Clear()
                msgs = Nothing


            End If

            'receive any new camera image from imgserver
            If Me.ConnectedToImageServer Then
                If Me.Config.ImageServerPort = 80 Then
                    'Dim htmlArgs As String = "?h=300&w=400&x=100&y=200&q=50"
                    Dim htmlArgs As String = "" 'full image
                    'Dim htmlGetRequest As String = "GET /i.png HTTP/1.1" 'only works if first byte is set to 137. Images black
                    'Dim htmlGetRequest As String = "GET /i.jpg HTTP/1.1" 'crashes
                    Dim htmlGetRequest As String = "GET /i.bmp HTTP/1.1"  'works. but coloured pixel on wrong location
                    'Dim htmlGetRequest As String = "GET /i.tif HTTP/1.1" 'crashes
                    Me._ImageServerConnection.Send(htmlGetRequest + htmlArgs)

                    If Me._ImageServerConnection.DataAvailable Then
                        Dim msg As String = Me._ImageServerConnection.Receive(640 * 480 * 3 + 240) 'HTML string. Header of has slightly more than 200 characters

                        'msg = "HTTP/1.1 200 OK  Content-Type: image/bmp  Content-Length: 76086  Accept-Ranges: none  Server: altURI/1.0  Cache-Control: no-cache  Last-Modified: Sun Aug 28 10:41:00 2011  Date: Sun Aug 28 10:41:00 2011 GMT    BM6)     6   (   ?   ?                     ...
                        'msg = "HTTP/1.1 200 OK  Content-Type: image/tiff  Content-Length: 36218  Accept-Ranges: none  Server: altURI/1.0  Cache-Control: no-cache  Last-Modified: Sun Aug 28 12:08:14 2011  Date: Sun Aug 28 12:08:14 2011 GMT    II* ??  ???????  ????A ?H ` ?...
                        'msg = "HTTP/1.1 200 OK  Content-Type: image/jpeg  Content-Length: 15172  Accept-Ranges: none  Server: altURI/1.0  Cache-Control: no-cache  Last-Modified: Sun Aug 28 10:21:41 2011  Date: Sun Aug 28 10:21:41 2011 GMT    ???? JFIF      ?? C ...
                        'msg = "HTTP/1.1 200 OK  Content-Type: image/png  Content-Length: 42429  Accept-Ranges: none  Server: altURI/1.0  Cache-Control: no-cache  Last-Modified: Fri Aug 26 12:14:31 2011  Date: Fri Aug 26 12:14:31 2011 GMT    ?PNG       IHDR   ?   ?   'yv?    IDATx?K...
                        'msg = "HTTP/1.1 404 Not Found  Content-Type: text/html  Content-Length: 154  Accept-Ranges: none  Server: altURI/1.0  Cache-Control: no-cache  Last-Modified: Sat Aug 27 07:26:04 2011  Date: Sat Aug 27 07:26:04 2011 GMT    <HTML><HEAD><TITLE>The file cannot be fou...

                        ' If msg.Contains("Content-Length: ") AndAlso msg.Contains("Accept-Ranges: ") AndAlso Not msg.Contains("The file cannot be found") Then
                        If msg.Contains("Content-Length: ") AndAlso msg.Contains("Accept-Ranges: ") Then
                            ' Get Length
                            Dim i As Integer = msg.IndexOf("Content-Length: ")
                            Dim j As Integer = msg.IndexOf("Accept-Ranges: ")
                            Dim part As String = msg.Substring(i + 16, j - i - 18) 'string.length("Content-Length: ")
                            Dim length As Integer = Integer.Parse(part)
                            Console.WriteLine(String.Format("Succesfull GetImage Request of length {0}", length))

                            If msg.Contains(" GMT") Then ' Data starts after time + 2x CRLF
                                i = msg.IndexOf(" GMT")
                                'part = msg.Substring(i + 8, msg.Length - i - 8)
                                'Dim png(msg.Length - i - 8) As Byte 'or one less?

                                'part = msg.Substring(i + 8, msg.Length - i - 8)
                                'Dim png(msg.Length - i - 9) As Byte 'or one less?

                                'part = msg.Substring(i + 6, msg.Length - i - 6) 'two bytes too long
                                'Dim png(msg.Length - i - 6) As Byte 'or one less?


                                part = msg.Substring(i + 8, msg.Length - i - 8) ' search for 137 80 78 71 13 10 26 10
                                Dim png(msg.Length - i - 8) As Byte 'or one less?

                                Dim encText As New System.Text.ASCIIEncoding() 'only needed for method (UTF8Encoding works also)

                                png = encText.GetBytes(part)
                                'png(0) = 137 ' was 63


                                Dim header(5) As Byte ' see http://athiri.cimds.ri.cmu.edu/foswiki/UsarSim/UpisInterface
                                header(0) = 6 'first free code after jpeg 
                                'header(0) = 1 'jpg with no compression (not checked)
                                'Dim length As Integer = CInt(bytes(1) * 2 ^ 24) + CInt(bytes(2) * 2 ^ 16) + CInt(bytes(3) * 2 ^ 8) + CInt(bytes(4) * 2 ^ 0)
                                header(1) = CByte(System.Math.Floor((png.Length / (2 ^ 24))))
                                header(2) = CByte(System.Math.Floor((png.Length - (header(1) * (2 ^ 24))) / (2 ^ 16)))
                                header(3) = CByte(System.Math.Floor((png.Length - (header(1) * (2 ^ 24)) - (header(2) * (2 ^ 16))) / (2 ^ 8)))
                                header(4) = CByte(System.Math.Floor((png.Length - (header(1) * (2 ^ 24)) - (header(2) * (2 ^ 16)) - (header(3) * (2 ^ 8))) / (2 ^ 0)))

                                Dim bytes(5 + png.Length - 1) As Byte

                                Array.ConstrainedCopy(header, 0, bytes, 0, 5)
                                header = Nothing
                                Array.ConstrainedCopy(png, 0, bytes, 5, png.Length)
                                png = Nothing


                                If Not ImageParser.ParseImageData(Me.Agent, bytes) Then
                                    Me.Error("Could not parse image data.", bytes.ToString)
                                Else
                                    Console.WriteLine("Succesfull Parsed HTML image")
                                End If



                                bytes = Nothing

                            End If
                        End If
                    End If
                Else 'Regular port 5003

                    If Me._ImageServerConnection.DataAvailable Then

                        bRecvImgs = True 'set the flag

                        ' get image data
                        'Dim bytes As Byte() = Me._ImageServerConnection.ReceiveImageData(921610) '921610 = 640 * 480 * 3 + 10), to prevent occassional Overflows
                        Dim bytes As Byte() = Me._ImageServerConnection.ReceiveImageData(320 * 240 * 3 + 10) 'size raw buffer

                        If Not IsNothing(bytes) AndAlso bytes.Length > 0 Then

                            If Not ImageParser.ParseImageData(Me.Agent, bytes) Then
                                Me.Error("Could not parse image data.", bytes.ToString)
                            End If
                            'Me._ImageServerConnection.SendAcknowledgement()
                        End If
                        bytes = Nothing 'Essential: This seems to be the memory leak that easily fills up 7Gb of Memory

                    End If


                End If


            End If

            'to avoid busy waiting
            If Not bSentCmds AndAlso Not bRecvMsgs AndAlso Not bRecvImgs Then

                'nothing happened this round, the agent may have crashed or 
                'perhaps it's the ComStation that does not have any sensors
                'avoid CPU overload while doing nothing: sleep 1 second

                'Threading.Thread.Sleep(TimeSpan.FromMilliseconds(100))
                'Threading.Thread.Sleep(TimeSpan.FromMilliseconds(33))
                'Threading.Thread.Sleep(TimeSpan.FromMilliseconds(337))
                Threading.Thread.Sleep(TimeSpan.FromMilliseconds(999))

                If Me.ConnectedToImageServer Then
                    Me._ImageServerConnection.SendAcknowledgement() 'Request an image
                End If
            Else
                'If Not IsNothing(Me.Agent) AndAlso Not IsNothing(Me.Agent.Name) Then
                '    If bSentCmds Then
                '        Console.WriteLine(String.Format("[LiveDriver]: {0} busy with sending commands", Me.Agent.Name))
                '    End If
                '    If bRecvMsgs Then
                '        Console.WriteLine(String.Format("[LiveDriver]: {0} busy with receiving messages", Me.Agent.Name))
                '    End If
                '    If bRecvImgs Then
                '        Console.WriteLine(String.Format("[LiveDriver]: {0} busy with receiving images", Me.Agent.Name))
                '    End If
                'End If
            End If



        End While

    End Sub

#End Region

#Region " Logging "

    Private logger As LogFileWriter

    Protected Sub OpenLog(ByVal logfilename As String)
        If Not IsNothing(Me.logger) Then Throw New InvalidOperationException("Log already opened")

        Dim fileInfo As New FileInfo(logfilename)
        If Not fileInfo.Directory.Exists Then
            fileInfo.Directory.Create()
        End If
        Me.LogFile = logfilename

        Me.logger = New LogFileWriter(logfilename)

    End Sub
    Protected Sub CloseLog()
        If Not IsNothing(Me.logger) Then
            Me.logger.Flush()
            Me.logger.Close()
            Me.logger.Dispose()
        End If
    End Sub

    Protected Overridable Sub [Error](ByVal msg As String, ByVal line As String)
        'comment the line before logging
        Me.Log(String.Format("{0} [ERROR] - {1}", My.Settings.CommentChar, line))
        Console.Error.WriteLine(msg & vbNewLine & vbTab & line)
    End Sub

    Protected Overridable Sub Log(ByVal line As String)
        If Not IsNothing(Me.logger) AndAlso line.Length > 3 Then 'To prevent flooded by many empty STA messages
            Me.logger.WriteLine(line)
        End If
    End Sub

#End Region

End Class
